/*
 * Serial driver module for PCTEL modems.
 *
 * 04/05: Added AC97 Codec ID reporting.
 *        Robert Thornburrow <robert.thornburrow@tsac.fsnet.co.uk>
 */

#include <linux/version.h>

#include <linux/module.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/sysrq.h>
#include <linux/delay.h>
#include <linux/device.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/serial_reg.h>
#include <linux/serial.h>

#include <asm/io.h>
#include <asm/irq.h>

#include "include/pctel.h"
#include "include/vuart.h"
#include "include/halwraps.h"
#include "ptserial-2.6.h"

#include "linmodem-2.6.h"

#define DRIVER_VERSION "v0.1"
#define DRIVER_AUTHOR  "Gustavo Sverzut Barbieri <barbieri@gmail.com>"
#define DRIVER_DESC    "PCTel serial modem driver"
#define DRIVER_LICENSE "GPL"

#define MY_NAME "pctel"

#ifdef PCTEL_DEBUG
extern int debug;

#define dbg(fmt, arg...)						    \
	do {								    \
		if (debug)						    \
			printk (KERN_DEBUG "%s(%d): %s: " fmt "\n",	    \
				MY_NAME , __LINE__, __FUNCTION__ , ## arg); \
	} while (0)
#else  /* PCTEL_DEBUG */
#define dbg(fmt, arg...) do { } while (0)
#endif /* PCTEL_DEBUG */

#define err(format, arg...) \
	printk(KERN_ERR "%s(%d): " format "\n", MY_NAME, __LINE__, ## arg)
#define info(format, arg...) \
	printk(KERN_INFO "%s(%d): " format "\n", MY_NAME, __LINE__, ## arg)
#define warn(format, arg...) \
	printk(KERN_WARNING "%s(%d): " format "\n", MY_NAME, __LINE__, ## arg)


static int country_code = 33;
module_param (country_code, int, 0);
MODULE_PARM_DESC(country_code, "Select a country code for the PCTEL modem.");


/* get_uart is array [10] of function pointers to functions taking void
   and returning unsigned char */
static unsigned char (*get_uart[10])(void) =
{
	get_uart_rx,
	get_uart_ier,
	get_uart_iir,
	get_uart_lcr,
	get_uart_mcr,
	get_uart_lsr,
	get_uart_msr,
	get_uart_scr,
	get_uart_dll,
	get_uart_dlm
};

/* put_uart is array [10] of function pointers to functions taking unsigned
   char and returning void */
static void (* ATTR_REGPARM() put_uart[10])(unsigned char) =
{
	put_uart_tx,
	put_uart_ier,
	put_uart_iir,
	put_uart_lcr,
	put_uart_mcr,
	put_uart_lsr,
	put_uart_msr,
	put_uart_scr,
	put_uart_dll,
	put_uart_dlm
};


static unsigned int pctel_serial_in(struct linmodem_port *p, int offset)
{
	if ((COM_Vlcr & UART_LCR_DLAB) && (offset <= UART_DLM))
		offset += (UART_SCR+1);

	return (*get_uart[offset])();
}

static void pctel_serial_out(struct linmodem_port *p, int offset, int value)
{
	if ((COM_Vlcr & UART_LCR_DLAB) && (offset <= UART_DLM))
		offset += (UART_SCR+1);

	(*put_uart[offset])(value);
}

#define pctel_serial_inp(port, offset)         pctel_serial_in(port, offset)
#define pctel_serial_outp(port, offset, value) pctel_serial_out(port, offset, value)

/* XXX: BEGIN HACK, remove these later! */
struct irq_info {
	spinlock_t		lock;
	struct list_head	*head;
};
#define is_real_interrupt(irq)	((irq) != 0)
static inline void
receive_chars(struct linmodem_port *p, int *status, struct pt_regs *regs)
{
	struct tty_struct *tty = p->port.info->tty;
	unsigned char ch, lsr = *status;
	int max_count = 256;
	char flag;

	dbg();

	do {
	    /* the next bit simply disappears in the 2.6.16 version of 8250.c */
#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16) )
		/* The following is not allowed by the tty layer and
		   unsafe. It should be fixed ASAP */
		if (unlikely(tty->flip.count >= TTY_FLIPBUF_SIZE)) {
			if(tty->low_latency)
				tty_flip_buffer_push(tty);
			/* If this failed then we will throw away the
			   bytes but must do so to clear interrupts */
		}
#endif

		ch = pctel_serial_inp(p, UART_RX);

		flag = TTY_NORMAL;
		p->port.icount.rx++;

		if (unlikely(lsr & (UART_LSR_BI | UART_LSR_PE |
				    UART_LSR_FE | UART_LSR_OE))) {
			/*
			 * For statistics only
			 */
			if (lsr & UART_LSR_BI) {
				lsr &= ~(UART_LSR_FE | UART_LSR_PE);
				p->port.icount.brk++;
				/*
				 * We do the SysRQ and SAK checking
				 * here because otherwise the break
				 * may get masked by ignore_status_mask
				 * or read_status_mask.
				 */
				if (uart_handle_break(&p->port))
					goto ignore_char;
			} else if (lsr & UART_LSR_PE)
				p->port.icount.parity++;
			else if (lsr & UART_LSR_FE)
				p->port.icount.frame++;
			if (lsr & UART_LSR_OE)
				p->port.icount.overrun++;

			/*
			 * Mask off conditions which should be ingored.
			 */
			lsr &= p->port.read_status_mask;

			if (lsr & UART_LSR_BI) {
				dbg("handling break...");
				flag = TTY_BREAK;
			} else if (lsr & UART_LSR_PE)
				flag = TTY_PARITY;
			else if (lsr & UART_LSR_FE)
				flag = TTY_FRAME;
		}

		/* this next part changes sometime before 2.6.13 in 8250.c */
#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,6,13) )
		if ((lsr & p->port.ignore_status_mask) == 0) {
			tty_insert_flip_char(tty, ch, flag);
		}
		if ((lsr & UART_LSR_OE) &&
		    tty->flip.count < TTY_FLIPBUF_SIZE) {
			/*
			 * Overrun is special, since it's reported
			 * immediately, and doesn't affect the current
			 * character.
			 */
			tty_insert_flip_char(tty, 0, TTY_OVERRUN);
		}
#else
		uart_insert_char(&p->port, lsr, UART_LSR_OE, ch, flag);
#endif

	ignore_char:
		lsr = pctel_serial_inp(p, UART_LSR);
	} while ((lsr & UART_LSR_DR) && (max_count-- > 0));
	tty_flip_buffer_push(tty);
	*status = lsr;
}

static inline void transmit_chars(struct linmodem_port *p)
{
	struct circ_buf *xmit = &p->port.info->xmit;
	/*int count;*/

	dbg();

	if (p->port.x_char) {
		pctel_serial_outp(p, UART_TX, p->port.x_char);
		p->port.icount.tx++;
		p->port.x_char = 0;
		return;
	}
	if (uart_circ_empty(xmit) || uart_tx_stopped(&p->port)) {
		linmodem_stop_tx(&p->port, 0);
		return;
	}

	/*count = p->tx_loadsz;
	do {
		pctel_serial_out(p, UART_TX, xmit->buf[xmit->tail]);
		xmit->tail = (xmit->tail + 1) & (UART_XMIT_SIZE - 1);
		p->port.icount.tx++;
		if (uart_circ_empty(xmit))
			break;
	} while (--count > 0);*/

	/* ### Jeff Trull's performance enhancement 19/07/2005 */
	while (uart_circ_chars_pending(xmit)
      	       && (pctel_serial_in(p, UART_LSR) & UART_LSR_THRE)) {
		pctel_serial_out(p, UART_TX, xmit->buf[xmit->tail]);
		xmit->tail++;
		if (xmit->tail >= UART_XMIT_SIZE) {
			xmit->tail = 0;
		}
		p->port.icount.tx++;
	}
	/* ### End Jeff Trull */

	if (uart_circ_chars_pending(xmit) < WAKEUP_CHARS)
		uart_write_wakeup(&p->port);

	if (uart_circ_empty(xmit))
		linmodem_stop_tx(&p->port, 0);
}

static inline void check_modem_status(struct linmodem_port *p)
{
	int status;

	dbg();

	status = pctel_serial_in(p, UART_MSR);

	if ((status & UART_MSR_ANY_DELTA) == 0)
		return;

	if (status & UART_MSR_TERI)
		p->port.icount.rng++;
	if (status & UART_MSR_DDSR)
		p->port.icount.dsr++;
	if (status & UART_MSR_DDCD)
		uart_handle_dcd_change(&p->port, status & UART_MSR_DCD);
	if (status & UART_MSR_DCTS)
		uart_handle_cts_change(&p->port, status & UART_MSR_CTS);

	wake_up_interruptible(&p->port.info->delta_msr_wait);
}

static inline void ptserial_clear_fifos(struct linmodem_port *p)
{
	dbg();

	if (p->capabilities & UART_CAP_FIFO) {
	        pctel_serial_outp(p, UART_FCR, UART_FCR_ENABLE_FIFO);
		pctel_serial_outp(p, UART_FCR, UART_FCR_ENABLE_FIFO |
				  UART_FCR_CLEAR_RCVR | UART_FCR_CLEAR_XMIT);
		pctel_serial_outp(p, UART_FCR, 0);
	}
}
static inline void
ptserial_handle_port(struct linmodem_port *p, struct pt_regs *regs)
{
	unsigned int status = pctel_serial_inp(p, UART_LSR);

	dbg();

	if (status & UART_LSR_DR) {
		receive_chars(p, &status, regs);
	}
	check_modem_status(p);
	if (status & UART_LSR_THRE){
		transmit_chars(p);
	}
}
/* XXX: END */

static irqreturn_t  pctel_interrupt(int irq, void *dev_id,
				    struct pt_regs *regs )
{
	static union i387_union i387;
	static unsigned long flags;
	static unsigned long cr0;
	struct irq_info *i = dev_id;
	struct linmodem_port *up;

	dbg();

	up = list_entry(i->head, struct linmodem_port, list);

	if (!HAL_DoInterrupt())
		return IRQ_RETVAL(0);

	spin_lock_irqsave(&up->port.lock, flags);

	/* make sure to update GlobalTimer */
	GlobalTimer = HAL_GetTimer();
	
	__asm__ __volatile__("mov %%cr0,%0 ; clts" : "=r" (cr0));

	__asm__ __volatile__("fnsave %0\n\t"
			     "fwait"
			     : "=m" (i387));

	dspMain();
	modem_main();

#ifdef AUDIO_ROUTING
	if ((audio_on_count >=5 ) || (ccount >= 10))
	{
		int zz = 0;
		audio_buffers.audio_on = 0;
		audio_buffers.last_filled = 0;

		for (zz=0; zz<SAMPLE_SIZE; zz++)
		{
			audio_buffers.buffer0[zz] = 0;
			audio_buffers.buffer1[zz] = 0;
			audio_buffers.buffer2[zz] = 0;
		}

		dsp_head = 0;
		dsp_tail = 0;

		for (zz=0; zz<480000; zz++)
			dsp_buffer[zz] = 0;

		audio_on_count = 0;
		ccount = 0;
	}
#endif /*  AUDIO_ROUTING */

	__asm__ __volatile__("frstor %0": :"m" (i387));

   __asm__ __volatile__("mov %0,%%cr0" : : "r" (cr0));

/*  spin_lock(&up->port.lock); */
	ptserial_handle_port(up, regs);
/*  spin_unlock(&up->port.lock); */

	spin_unlock_irqrestore(&up->port.lock, flags);

	return IRQ_RETVAL(1);
}

static void pctel_set_mctrl(struct linmodem_port *p, unsigned int mctrl)
{
	unsigned char mcr = 0;

	dbg();

	if (mctrl & TIOCM_RTS)
		mcr |= UART_MCR_RTS;
	if (mctrl & TIOCM_DTR)
		mcr |= UART_MCR_DTR;
	if (mctrl & TIOCM_OUT1)
		mcr |= UART_MCR_OUT1;
	if (mctrl & TIOCM_OUT2)
		mcr |= UART_MCR_OUT2;
	if (mctrl & TIOCM_LOOP)
		mcr |= UART_MCR_LOOP;

	mcr = mcr | p->mcr;


	pctel_serial_out(p, UART_MCR, mcr);
}

static void pctel_autoconfig(struct linmodem_port *p, unsigned int probeflags)
{
	dbg();

	if (p->port.line == PCTEL_TTY_LINE)
	{
		p->port.type     = PORT_PCTEL;
		p->port.fifosize = p->config.fifo_size;

		return;
	}
}

static int pctel_startup(struct linmodem_port *p)
{
	unsigned long flags;
	int retval;

	dbg();

	p->capabilities = p->config.flags;
	p->mcr = 0;

	info("PCTel initialization. Country code is %d.\n", country_code);

	/* increment pctel usage */
	/*Pctel_module_init();*/

	PctelInitVUartVars();
	PctelInitCtrlVars(country_code);
	/*
	 * Clear the FIFO buffers and disable them.
	 * (they will be reeanbled in set_termios())
	 */
	ptserial_clear_fifos(p);

	/*
	 * Clear the interrupt registers.
	 */
	pctel_serial_inp(p, UART_LSR);
	pctel_serial_inp(p, UART_RX);
	pctel_serial_inp(p, UART_IIR);
	pctel_serial_inp(p, UART_MSR);

	/*
	 * At this point, there's no way the LSR could still be 0xff;
	 * if it is, then bail out, because there's likely no UART
	 * here.
	 */
	if (!(p->port.flags & UPF_BUGGY_UART) &&
	    (pctel_serial_inp(p, UART_LSR) == 0xff)) {
		info("%s%d: LSR safety check engaged!",
		     p->reg.dev_name, p->port.line);
		return -ENODEV;
	}


	/*
	 * If the "interrupt" for this port doesn't correspond with any
	 * hardware interrupt, we use a timer-based system.  The original
	 * driver used to do this with IRQ0.
	 */
	if (!is_real_interrupt(p->port.irq)) {
		unsigned int timeout = p->port.timeout;

		timeout = timeout > 6 ? (timeout / 2 - 2) : 1;

		p->timer.data = (unsigned long)up;
		mod_timer(&p->timer, jiffies + timeout);
	} else {
		retval = serial_link_irq_chain(p);
		if (retval)
			return retval;
	}

	/*
	 * Now, initialize the UART
	 */
	pctel_serial_outp(p, UART_LCR, UART_LCR_WLEN8);

	spin_lock_irqsave(&p->port.lock, flags);
	/*
	 * Most PC uarts need OUT2 raised to enable interrupts.
	 */
	if (is_real_interrupt(p->port.irq))
		p->port.mctrl |= TIOCM_OUT2;

	pctel_set_mctrl(p, p->port.mctrl);
	spin_unlock_irqrestore(&p->port.lock, flags);

	/*
	 * Finally, enable interrupts.  Note: Modem status interrupts
	 * are set via set_termios(), which will be occurring imminently
	 * anyway, so we don't enable them here.
	 */
	p->ier = UART_IER_RLSI | UART_IER_RDI ;
	pctel_serial_outp(p, UART_IER, p->ier);

#if defined (HAL_I8XX) || defined (HAL_VIA686A)
	/* request_region(pctel_region_0(p->port), 256, "PCTel AC97(mixer)"); */
	/* request_region(pctel_region_1(p->port), 128, "PCTel AC97(modem)"); */
#elif defined (HAL_CM8738)
	/* request_region(pctel_region_0(p->port), 32, "PCTel CM8738"); */
#else
	request_region(p->port.iobase, 8, "PCTel PCT789");
#endif

	HAL_Init( p->port.iobase, (int) pctel_region_1(p->port) );

#if 0 /*defined (HAL_I8XX) || defined (HAL_VIA686A)*/
	/*
	 * Report the kind of codec found - i8xxhal.s, sishal.a and
	 * via686ahal.a use gwCodecID to adjust the code to handle different
	 * codecs.  These are:
	 *
	 *   0x0  ------  -------------  UNSUPPORTED
	 *   0x1  DT2-49  0x4454:0x3231  PCTel Delta
	 *   0x1  DT2-50  0x4454:0x3230  PCTel Delta
	 *   0x2  SIL17   0x5349:0x4c11  PCTel/Silicon Lab Stinger
	 *   0x2  SIL33   0x5349:0x4c21  PCTel/Silicon Lab New
	 *   0x3  TRA09   0x5452:0x4109  TriTech
	 */
	extern unsigned short gwCodecID;

#if defined(INTEL)
	/*
	 * The Intel i8xx HAL has been modified to return the codec vendor id.
	 * This is decoded as AAAnn or xxxx:xxxx depending upon whether the first
	 * word is a pair of valid ASCII characters or not.
	 */
	extern unsigned short gwVendorID1, gwVendorID2;

	if( gwVendorID1 && gwVendorID2 )
	{
		if( gwVendorID1 & 0x8080 ) {
			if( gwCodecID )
				info("Found codec 0x%04x:0x%04x of type %x",
				     gwVendorID1, gwVendorID2, gwCodecID);
			else
				info("Found unsupported codec 0x%04x:0x%04x",
				     gwVendorID1, gwVendorID2);
		} else {
			char buf[7];
            
			buf[0] = (gwVendorID1 >> 8);
			buf[1] = (gwVendorID1 & 0xff);
			buf[2] = (gwVendorID2 >> 8);
    		
			snprintf( buf+3, 4, "%d", gwVendorID2 & 0xff );
    		
			if( gwCodecID )
				info("Found codec %s of type %x\n",
				     buf, gwCodecID);
			else
				info("Found unsupported codec %s\n", buf);
		}
	}
#else
	if( gwCodecID )
		info("Found codec of type %x", gwCodecID);
	else
		info("Found unsupported codec");
#endif /* INTEL */
#endif /* HAL_I8XX || HAL_VIA686A */

	/*
	 * And clear the interrupt registers again for luck.
	 */
	pctel_serial_inp(p, UART_LSR);
	pctel_serial_inp(p, UART_RX);
	pctel_serial_inp(p, UART_IIR);
	pctel_serial_inp(p, UART_MSR);

	return 0;
}

static void pctel_shutdown(struct linmodem_port *p)
{
	unsigned long flags;

	dbg();

	HAL_Deinit(p->port.iobase);

	/* free pctel usage */
	/* Pctel_module_free(); */

#if defined (HAL_I8XX) || defined (HAL_VIA686A)
        /* release_region(pctel_region_0(p->port), 256); */
	/* release_region(pctel_region_1(p->port), 128); */
#elif defined (HAL_CM8738)
	/* release_region(pctel_region_0(p->port), 32); */
#else
	release_region(p->port.iobase, 8);
#endif

	/*
	 * Disable interrupts from this port
	 */
	p->ier = 0;
	pctel_serial_outp(p, UART_IER, 0);

	spin_lock_irqsave(&p->port.lock, flags);
	if (p->port.flags & UPF_FOURPORT) {
		/* reset interrupts on the AST Fourport board */
		inb((p->port.iobase & 0xfe0) | 0x1f);
		p->port.mctrl |= TIOCM_OUT1;
	} else
		p->port.mctrl &= ~TIOCM_OUT2;

	pctel_set_mctrl(p, p->port.mctrl);
	spin_unlock_irqrestore(&p->port.lock, flags);

	/*
	 * Disable break condition and FIFOs
	 */
	pctel_serial_out(p, UART_LCR,
			 pctel_serial_inp(p, UART_LCR) & ~UART_LCR_SBC);
	ptserial_clear_fifos(p);

	/*
	 * Read data port to reset things, and then unlink from
	 * the IRQ chain.
	 */
	pctel_serial_in(p, UART_RX);

	if (!is_real_interrupt(p->port.irq))
		del_timer_sync(&p->timer);
	else
		serial_unlink_irq_chain(p);
}




struct linmodem_ops pctel_ops = {
	.serial_in  = pctel_serial_in,
	.serial_out = pctel_serial_out,
	.interrupt  = pctel_interrupt,
	.set_mctrl  = pctel_set_mctrl,
	.autoconfig = pctel_autoconfig,
	.startup    = pctel_startup,
	.shutdown   = pctel_shutdown,
};

struct linmodem_port pctel_port = {
	.port = {},
	.ops  = &pctel_ops,
	.config = {
		.name      = "PCTel",
		.fifo_size = 128,
		.flags     = 0,
	},
	.reg = {
		.owner = THIS_MODULE,
		.driver_name = "pctel",
/*		.devfs_name  = "ttpctel/", */
		.dev_name    = "ttyS_PCTEL",
		.major       = PCTEL_MAJOR,
		.minor       = PCTEL_MINOR,
		.nr          = 1,
	},
};


int pctel_init(void)
{
	dbg();

    info( MY_NAME " " DRIVER_VERSION " loaded" );

	return linmodem_new(&pctel_port);
}

void pctel_exit(void)
{
	dbg();

    info( MY_NAME " " DRIVER_VERSION " unloaded" );

	linmodem_del(&pctel_port);
}



int pctel_new(struct uart_port *port)
{
	struct uart_port *p = &(pctel_port.port);
	dbg();

	p->iobase   = port->iobase;
	p->membase  = port->membase;
	p->irq      = port->irq;
	p->uartclk  = port->uartclk;
	p->fifosize = port->fifosize;
	p->regshift = port->regshift;
	p->iotype   = port->iotype;
	p->flags    = port->flags | UPF_BOOT_AUTOCONF;
	p->mapbase  = port->mapbase;

	return linmodem_hw_new((struct uart_port*)&pctel_port);
}

EXPORT_SYMBOL(pctel_new);
